package main

import (
	"errors"
	"fmt"
	"launchpad.net/lpad"
	"strings"
)

func loginLaunchpad() (*lpad.Root, error) {
	logf("Authenticating in Launchpad...")
	root, err := lpad.Login(lpad.Production, &lpad.ConsoleOAuth{})
	if err != nil {
		return nil, fmt.Errorf("Can't authenticate in Launchpad: %v", err)
	}
	return root, nil
}

func loadMe(root *lpad.Root) (*lpad.Person, error) {
	logf("Looking up your user details in Launchpad...")
	return root.Me()
}

var stop = errors.New("Stop marker")

func loadProject(root *lpad.Root, name string) (*lpad.Project, *lpad.Milestone, error) {
	logf("Loading data for project %q...", name)
	project, err := root.Project(name)
	if err != nil {
		return nil, nil, fmt.Errorf("Failed to load data for project %q: %v", name, err)
	}
	logf("Loading active milestones for project %q...", name)
	list, err := project.ActiveMilestones()
	if err != nil {
		return nil, nil, fmt.Errorf("Failed to load active milestones for project %q: %v", name, err)
	}
	var ms *lpad.Milestone
	if list.TotalSize() > 0 {
		err := list.For(func(inner *lpad.Milestone) error {
			ms = inner
			return stop
		})
		if err != nil && err != stop {
			return nil, nil, fmt.Errorf("Failed to iterate over active milestones: %v", err)
		}
		logf("Next active milestone for project %q is %q.", name, ms.Name())
	} else {
		logf("Project %q has no active milestones.", name)
	}
	return project, ms, nil
}

func loadMember(root *lpad.Root, username string) (lpad.Member, error) {
	logf("Loading data for member %q...", username)
	member, err := root.Member(username)
	if err != nil {
		return nil, fmt.Errorf("Failed to load data for member %q: %v", username, err)
	}
	return member, nil
}

func loadMemberAddress(member lpad.Member) (string, error) {
	person, ok := member.(*lpad.Person)
	if !ok {
		return "", fmt.Errorf("Launchpad member %q is a team.", member.Name())
	}
	name := person.DisplayName()
	logf("Loading preferred email for %q...", name)
	email, err := person.PreferredEmail()
	if err != nil {
		if err == lpad.ErrNotFound {
			// User disabled access to email addresses.
			return name, nil
		}
		return "", fmt.Errorf("Failed to load person's preferred email: %v", err)
	}
	return fmt.Sprintf("%s <%s>", name, email), nil
}

func loadBranch(root *lpad.Root, id string) (*lpad.Branch, error) {
	logf("Loading data for branch %s...", id)
	branch, err := root.Branch(id)
	if err != nil {
		return nil, fmt.Errorf("Failed to load data for branch %s: %v", id, err)
	}
	return branch, nil
}

func loadBug(root *lpad.Root, id int) (*lpad.Bug, error) {
	logf("Loading data for bug %d...", id)
	bug, err := root.Bug(id)
	if err != nil {
		return nil, fmt.Errorf("Failed to load data for bug %d: %v", id, err)
	}
	return bug, nil
}

func loadBlueprint(root *lpad.Root, target lpad.BlueprintTarget, name string) (*lpad.Blueprint, error) {
	logf("Loading data for blueprint %q...", name)
	bp, err := root.Blueprint(target, name)
	if err != nil {
		return nil, fmt.Errorf("Failed to load data for blueprint %q: %v", bp, err)
	}
	return bp, nil
}

func createBug(root *lpad.Root, stub *lpad.BugStub) (*lpad.Bug, error) {
	logf("Creating bug...")
	bug, err := root.CreateBug(stub)
	if err != nil {
		return nil, fmt.Errorf("Failed to create bug: %v", err)
	}
	logf("Created bug: %s", bug.WebPage())
	return bug, nil
}

func linkBugBranch(bug *lpad.Bug, branch *lpad.Branch) error {
	logf("Linking branch %s to bug %d...", branch.Id(), bug.Id())
	if err := bug.LinkBranch(branch); err != nil {
		return fmt.Errorf("Failed to link branch to bug: %v", err)
	}
	return nil
}

func linkBlueprintBranch(bp *lpad.Blueprint, branch *lpad.Branch) error {
	logf("Linking blueprint %q to branch %s...", bp.Name(), branch.Id())
	if err := bp.LinkBranch(branch); err != nil {
		return fmt.Errorf("Failed to link blueprint to branch: %v", err)
	}
	return nil
}

func linkBlueprintBug(bp *lpad.Blueprint, bug *lpad.Bug) error {
	logf("Linking bp %q to bug %d...", bp.Name(), bug.Id())
	if err := bp.LinkBug(bug); err != nil {
		return fmt.Errorf("Failed to link bp to bug: %v", err)
	}
	return nil
}

func setInProgress(bug *lpad.Bug, milestone *lpad.Milestone, person *lpad.Person) error {
	logf("Loading bug %d task...", bug.Id())
	var task *lpad.BugTask
	tasks, err := bug.Tasks()
	if err == nil {
		err = tasks.For(func(t *lpad.BugTask) error {
			task = t
			return stop
		})
	}
	if err != nil && err != stop {
		return fmt.Errorf("Failed to retrieve bug %d task: %v", bug.Id(), err)
	}
	logf("Updating bug %d task status, milestone, and assignee...", bug.Id())
	task.SetStatus(lpad.StInProgress)
	task.SetAssignee(person)
	if milestone != nil {
		task.SetMilestone(milestone)
	}
	if err = task.Patch(); err != nil {
		return fmt.Errorf("ERROR: Failed to update bug task: %v", err)
	}
	return nil
}

var notFound = errors.New("not found")

func loadProposal(source, target *lpad.Branch) (mp *lpad.MergeProposal, err error) {
	logf("Checking if merge proposal already exists...")
	mps, err := source.LandingTargets()
	if err != nil {
		return nil, fmt.Errorf("Failed to obtain merge proposals: %v", err)
	}
	err = mps.For(func(inner *lpad.MergeProposal) error {
		branch, err := inner.Target()
		if err != nil {
			return fmt.Errorf("Failed to get target branch: %v", err)
		}
		if branch.Id() == target.Id() {
			mp = inner
			return stop
		}
		return nil
	})
	if err != nil && err != stop {
		return nil, err
	}
	if mp == nil {
		logf("Merge proposal not found.")
		return nil, notFound
	}
	logf("Merge proposal %s => %s found: %s", source.Id(), target.Id(), mp.WebPage())
	return mp, nil
}

func inferTarget(source *lpad.Branch) (target *lpad.Branch, mp *lpad.MergeProposal, err error) {
	logf("Looking for merge proposal...")
	mps, err := source.LandingTargets()
	if err != nil {
		return nil, nil, fmt.Errorf("Failed to obtain merge proposals: %v", err)
	}
	err = mps.For(func(inner *lpad.MergeProposal) error {
		if mp != nil {
			return fmt.Errorf("Multiple merge proposals for source branch.")
		}
		target, err = inner.Target()
		if err != nil {
			return fmt.Errorf("Failed to get target branch: %v", err)
		}
		mp = inner
		return nil
	})
	if err != nil {
		return nil, nil, err
	}
	if target == nil {
		return nil, nil, fmt.Errorf("Merge proposal for %s not found.", source.Id())
	}
	logf("Merge proposal %s => %s found: %s", source.Id(), target.Id(), mp.WebPage())
	return target, mp, nil
}

func loadFocusBranch(project *lpad.Project) (*lpad.Branch, error) {
	logf("Loading focus series for project %q...", project.Name())
	series, err := project.FocusSeries()
	if err != nil {
		return nil, fmt.Errorf("Failed to load focus series for %q: %v", project.Name(), err)
	}
	logf("Focus series for project %q is %q.", project.Name(), series.Name())
	logf("Loading target branch for series %q...", series.Name())
	target, err := series.Branch()
	if err != nil {
		return nil, fmt.Errorf("Failed to load target branch for series %q: %v", series.Name(), err)
	}
	logf("Target branch for series %q is %s", series.Name(), target.Id())
	return target, nil
}

func proposeMerge(project *lpad.Project, prereq, source, target *lpad.Branch, changelog *Log, prepare bool) (*lpad.MergeProposal, error) {
	logf("Creating merge proposal: %s => %s", source.Id(), target.Id())
	stub := &lpad.MergeStub{
		Description: changelog.Summary + "\n\n" + changelog.Description,
		Target:      target,
		PreReq:      prereq,
		NeedsReview: !prepare,
	}
	mp, err := source.ProposeMerge(stub)
	if err != nil {
		return nil, fmt.Errorf("Failed to create merge proposal: %v", err)
	}
	logf("Merge proposal successfully created: %s", mp.WebPage())
	return mp, nil
}

func isMerged(source *lpad.Branch) (merged bool, mp *lpad.MergeProposal, err error) {
	mps, err := source.LandingTargets()
	if failed(err) {
		return
	}
	found := errors.New("marker")
	err = mps.For(func(proposal *lpad.MergeProposal) error {
		if proposal.Status() == lpad.StMerged {
			mp = proposal
			return found
		}
		return nil
	})
	switch err {
	case nil:
		return false, mp, nil
	case found:
		return true, mp, nil
	}
	return false, nil, err
}

func setProposalLog(mp *lpad.MergeProposal, log *Log) error {
	logf("Updating merge proposal log...")
	desc := fmt.Sprintf("%s\n\n%s", log.Summary, log.Description)
	if log.CodeReviewId != 0 {
		desc = fmt.Sprintf("%s\n\nhttps://codereview.appspot.com/%d/\n", desc, log.CodeReviewId)
	}
	mp.SetDescription(desc)
	err := mp.Patch()
	if err != nil {
		return fmt.Errorf("Failed to update merge proposal log: %v", err)
	}
	return nil
}

func setProposalStatus(mp *lpad.MergeProposal, status lpad.MergeProposalStatus) error {
	if mp.Status() == status {
		logf("Merge proposal already set to %s.", status)
		return nil
	}
	logf("Updating merge proposal status to %s...", status)
	err := mp.SetStatus(status)
	if err != nil {
		return fmt.Errorf("Failed to update merge proposal status: %v", err)
	}
	return nil
}

func addProposalComment(mp *lpad.MergeProposal, message string) error {
	logf("Adding comment to merge proposal...")
	err := mp.AddComment("", message, lpad.VoteNone, "")
	if err != nil {
		return fmt.Errorf("Error adding comment to merge proposal: %v", err)
	}
	return nil
}

func removeMp(list []string) []string {
	filtered := make([]string, 0, len(list))
	for _, s := range list {
		if !strings.HasSuffix(s, "@code.launchpad.net") {
			filtered = append(filtered, s)
		}
	}
	return filtered
}

func justMp(list []string) []string {
	filtered := make([]string, 0, len(list))
	for _, s := range list {
		if strings.HasSuffix(s, "@code.launchpad.net") {
			filtered = append(filtered, s)
		}
	}
	return filtered
}
